#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/serial.h>
#include <string.h>
#include <asm-generic/ioctls.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/ioctl.h>

#define F81504_CONFIG_SYSFS "/sys/class/tty/%s/device/config"
#define F81504_CONFIG_SYSFS_IOPORT "/sys/class/tty/%s/port"
#define F81504_CONFIG_SYSFS_MEM_BASE "/sys/class/tty/%s/iomem_base"
#define F81504_SERIAL_SYSFS_IOPORT "/proc/tty/driver/serial"

#define VERSION "v20241025"

enum error {
	error_no_error = 0,

	error_parameter,
	error_open_failed,
	error_not_f81504,
	error_change_mode,
	error_read_id,
	error_read_io,
	error_read_pci,
	error_write_pci,
	error_io_not_found,
};

static char *m_cStringArray[] = {
	(char *)-1,
	"eModeRS232",
	"eModeRS485 - DE#/RE (TX Enable with RTS Low)",
	"eModeRS485_1 - DE/RE# (TX Enable with RTS High)",
	(char *)-1,
	NULL,
};

void usage(char *argv[])
{
	int i = 0;
	fprintf(stderr, "%s\n", argv[0]);
	fprintf(stderr, "%s <port> <mode>\n\n", argv[0]);
	fprintf(stderr, "\nUart Mode table:\n");

	for (i = 0; m_cStringArray[i]; ++i) {
		if (m_cStringArray[i] == (char *)-1)
			continue;
		fprintf(stderr, "%i --> %s\n", i, m_cStringArray[i]);
	}

	fprintf(stderr, "\n");
}

int is_f81504(int fd)
{
	int len;
	int tmp;
	char buf[5];

	memset(buf, 0, sizeof(buf));

	lseek(fd, 0, SEEK_SET);

	len = read(fd, buf, sizeof(buf));
	if (!len)
		return error_read_id;

	for (len = 0; len < 4; ++len)
		printf("%x", buf[len]);
	printf("\n");

	if ((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x11))
		return error_not_f81504;

	switch (buf[2]) {
	case 0x04:
	case 0x08:
	case 0x12:
		return error_no_error;
	}

	return error_not_f81504;
}

int is_f81504a(int fd)
{
	int len;
	int tmp;
	char buf[5];

	memset(buf, 0, sizeof(buf));

	lseek(fd, 0, SEEK_SET);

	len = read(fd, buf, sizeof(buf));
	if (!len)
		return error_read_id;

	for (len = 0; len < 4; ++len)
		printf("%x", buf[len]);
	printf("\n");

	if ((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x13))
		return error_not_f81504;

	switch (buf[2]) {
	case 0x04:
	case 0x08:
	case 0x12:
		return error_no_error;
	}

	return error_not_f81504;
}

int change_mode_f81504(int fd, int mode, int io_addr)
{
	unsigned char buf[4];
	int read_io;
	int len;
	int i;

	for (i = 0; i < 12; ++i) {
		memset(buf, 0, sizeof(buf));
		lseek(fd, 0x40 + 0x8 * i + 4, SEEK_SET);

		len = read(fd, buf, 2);
		if (len != 2)
			return error_read_pci;

		read_io = (buf[1] << 8) | buf[0];

		if (io_addr == read_io)
			break;
	}

	if (i >= 12)
		return error_io_not_found;

	memset(buf, 0, sizeof(buf));
	lseek(fd, 0x40 + 0x8 * i + 7, SEEK_SET);

	len = read(fd, buf, 1);
	if (len != 1)
		return error_read_pci;

	switch (mode) {
	case 1:
		buf[0] = 0x01;
		break;
	case 2:
		buf[0] = 0x31;
		break;
	case 3:
		buf[0] = 0x11;
		break;
	default:
		return error_parameter;
	}

	lseek(fd, 0x40 + 0x8 * i + 7, SEEK_SET);
	len = write(fd, buf, 1);
	if (len != 1)
		return error_write_pci;

	return error_no_error;
}

int change_mode_f81504a(int fd, int mode, int io_addr)
{
	uint32_t res = 0, idx;
	uint8_t tmp;
	int read_io;
	int len;
	int i;

	lseek(fd, 0x10, SEEK_SET);

	len = read(fd, &res, 4);
	if (len != 4)
		return error_read_pci;

	if (io_addr < res || ((io_addr - res) % 8)) {
		printf("%s: res err: %x %x\n", __func__, io_addr, res);
		return error_read_io;
	}

	switch (mode) {
	case 1:
		tmp = 0x01;
		break;
	case 2:
		tmp = 0x31;
		break;
	case 3:
		tmp = 0x11;
		break;
	default:
		return error_parameter;
	}

	idx = (io_addr - res) / 8;
	lseek(fd, 0x2a0 + 8 * idx + 7, SEEK_SET);
	len = write(fd, &tmp, 1);
	if (len != 1)
		return error_read_pci;

	printf("%s: tmp: %x\n", __func__, tmp);

	return error_no_error;
}

int read_resource(char *tty)
{
	char buf[128];
	int len, tty_idx, cur_idx;
	int fd;
	FILE *file;
	int io_addr = -1;
	char *pos, *pos_space;

	do {
		sprintf(buf, F81504_CONFIG_SYSFS_IOPORT, tty);

		fd = open(buf, O_RDONLY);
		if (fd < 0) {
			break;
		}

		memset(buf, 0, sizeof(buf));
		len = read(fd, buf, sizeof(buf));
		if (!len) {
			close(fd);
			break;
		}

		io_addr = strtol(buf, NULL, 16);
		if (io_addr == 0) {
			close(fd);
			break;
		}

		close(fd);
		return io_addr;
	} while (0);

	do {
		sprintf(buf, F81504_CONFIG_SYSFS_MEM_BASE, tty);

		fd = open(buf, O_RDONLY);
		if (fd < 0) {
			break;
		}

		memset(buf, 0, sizeof(buf));
		len = read(fd, buf, sizeof(buf));
		if (!len) {
			close(fd);
			break;
		}

		io_addr = strtol(buf, NULL, 16);
		if (io_addr == 0) {
			close(fd);
			break;
		}

		close(fd);
		return io_addr;
	} while (0);

	sprintf(buf, F81504_SERIAL_SYSFS_IOPORT);
	file = fopen(buf, "r");
	if (!file)
		return error_open_failed;

	tty_idx = strtol(&tty[4], NULL, 10);

	io_addr = error_read_io;
	while (fgets(buf, sizeof(buf), file)) {
		if (buf[0] < '0' || buf[0] > '9')
			continue;

		cur_idx = strtol(buf, NULL, 10);

		if (tty_idx != cur_idx)
			continue;

		//printf("%s", buf);

		pos = strstr(buf, "port:");
		pos_space = strstr(&buf[pos - buf + 5], " ");
		*pos_space = 0;

		//printf("pos: %d\n", pos - buf);
		//printf("%s\n", &buf[pos - buf + 5]);

		io_addr = strtol(&buf[pos - buf + 5], NULL, 16);
		break;
	}

	fclose(file);
	return io_addr;
}

int rs485_api(char *dev, int mode)
{
	struct serial_rs485 rs485conf;
	uint8_t buf[128];
	int fd;

	sprintf(buf, "/dev/%s", dev);
	printf("Target:%s\n", buf);

	fd = open(buf, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "open error %d\n", __LINE__);
		return error_parameter;
	}

	if (ioctl(fd, TIOCGRS485, &rs485conf) < 0) {
		//printf("Error: TIOCGRS485 ioctl not supported.\n");
		close(fd);
		return error_parameter;
	}

#if 0
	if (rs485conf.flags & SER_RS485_ENABLED) {
		if (rs485conf.flags & SER_RS485_RTS_ON_SEND)
			printf("current mode: 3, eModeRS485_1\n");
		else
			printf("current mode: 2, eModeRS485\n");
	} else {
		printf("current mode: 1, eModeRS232\n");
	}
#endif

	switch (mode) {
	case 1:
		rs485conf.flags = SER_RS485_RTS_AFTER_SEND;
		break;
	case 2:
		rs485conf.flags = SER_RS485_ENABLED | SER_RS485_RTS_AFTER_SEND;
		break;
	case 3:
		rs485conf.flags = SER_RS485_ENABLED | SER_RS485_RTS_ON_SEND;
		break;
	default:
		return error_parameter;
	}

	if (ioctl(fd, TIOCSRS485, &rs485conf) < 0) {
		printf("Error: TIOCSRS485 ioctl not supported.\n");
		close(fd);
		return error_parameter;
	}

	printf("change to mode: %d, %s\n", mode, m_cStringArray[mode]);

	close(fd);
	return 0;
}

int main(int argc, char *argv[])
{
	int fd = -1;
	int length = 0;
	int i = 0;
	int ret = 0;
	char buf[128];
	int status;
	struct serial_rs485 rs485conf;

	printf("VERSION: %s\n", VERSION);

#if 0
	read_resource("ttyS10");
	return 0;
#endif

	if (argc < 3) {
		usage(argv);
		return error_parameter;
	}

	i = atoi(argv[2]);

	if (m_cStringArray[i] == (char *)-1) {
		usage(argv);
		return error_parameter;
	}

	ret = rs485_api(argv[1], i);
	if (ret == 0) {
		printf("rs485 api ok\n");
		return 0; //success
	}

	sprintf(buf, F81504_CONFIG_SYSFS, argv[1]);
	printf("Target:%s\n", buf);

	fd = open(buf, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "open error %d\n", __LINE__);
		return error_open_failed;
	}

	if (!is_f81504a(fd)) {
		status = change_mode_f81504a(fd, i, read_resource(argv[1]));
		if (status) {
			fprintf(stderr, "change_mode failed\n");
			close(fd);
			return status;
		}

		printf("change to mode: %s\n", m_cStringArray[i]);

	} else if (!is_f81504(fd)) {
		status = change_mode_f81504(fd, i, read_resource(argv[1]));
		if (status) {
			fprintf(stderr, "change_mode failed\n");
			close(fd);
			return status;
		}

		printf("change to mode: %s\n", m_cStringArray[i]);
	} else {
		status = error_not_f81504;

		printf("can't found any Fintek PCIE to serial product\n");
	}

	close(fd);

	return status;
}
