/*
    Fintek F81504 / F81508 / F81512

    Tested on
        Linux Kernel 3.4.5
*/

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/tty.h>
#include <linux/serial_reg.h>
#include <linux/serial_core.h>
#include <linux/8250_pci.h>
#include <linux/bitops.h>
#include <linux/serial_8250.h>
#include <linux/version.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <linux/mutex.h>

#include <linux/serial_8250.h>
#include <linux/serial_reg.h>
#include <linux/dmaengine.h>
#include <linux/gpio.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 0, 0)
#include <linux/gpio/driver.h>
#endif

#define F81504_DRIVER_VER "1.0.17-20241030"

#define FINTEK_DEBUG 0
#define EXTRA_GPIO_CHANGE_MODE 1

#define LOG_MESSAGE(x, y, ...)                      \
	do {                                        \
		if (FINTEK_DEBUG)                   \
			printk(x y, ##__VA_ARGS__); \
	} while (0)

#define MAX_FINTEK_SERIAL_PORT 12
#define MAX_FINTEK_SERIAL_GPIO 6
#define MULTI_FUNCTION_PORT_FLAG -1

#define GPIO_MODE_OPENDRAIN
#define AER_FORCE_CHECK

#define NO_GPIO_DEV 0xff
//#define F81504_I2C_CONF	0xa08
#define F81504_I2C_CONF 0xa00

#if 0
#define FORCE_INTO_GPIO_MODE
#endif

//#undef CONFIG_GPIOLIB

#define DRIVER_NAME "Fintek F81504/F81508/F81512"
#define GPIO_DISPLAY_NAME "F81504-Series-GPIO"

static unsigned int disable_port = 0;
module_param(disable_port, uint, S_IRUGO);
MODULE_PARM_DESC(disable_port, "disable_port, using bitwise");

static u8 skip_port[MAX_FINTEK_SERIAL_GPIO] = { 2, 3, 8, 9, 10, 11 };

typedef enum {
	transceiver_type_f81439,
	transceiver_type_f81435,
	transceiver_type_f81433,
	transceiver_type_invalid,
} transceiver_type;

struct f81504_priv {
#ifdef CONFIG_GPIOLIB
	struct gpio_chip f81504_gpio_chip[6];
#endif
	struct mutex locker;
	int line_data[MAX_FINTEK_SERIAL_PORT];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	struct serial_rs485 rs485[MAX_FINTEK_SERIAL_PORT];
#endif
	u32 gpio_ioaddress;
	u8 multi_function;

#if EXTRA_GPIO_CHANGE_MODE
	bool is_i2c_mem;
	unsigned char __iomem *i2c_mem;
	int gpio_dev_id;
	int gpio_mode[MAX_FINTEK_SERIAL_PORT];
	transceiver_type transceiver_type[MAX_FINTEK_SERIAL_PORT];
#endif

	u8 rs485_reg[MAX_FINTEK_SERIAL_PORT];
};

static u32 baudrate_table[] = { 115200, 1500000, 1152000, 921600 };
static u8 clock_table[] = { 0x00, 0x80, 0x40, 0xc0 }; /* clock + fifo enable */

#ifdef CONFIG_GPIOLIB
static int f81504_check_valid_gpio(struct gpio_chip *chip)
{
	int idx;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	for (idx = 0; idx < MAX_FINTEK_SERIAL_GPIO; ++idx) {
		if (priv->f81504_gpio_chip[idx].label &&
		    (!strcmp(priv->f81504_gpio_chip[idx].label, chip->label)))
			break;
	}

	if (idx >= MAX_FINTEK_SERIAL_GPIO)
		return -EINVAL;

	return idx;
}

static int f81504_gpio_get(struct gpio_chip *chip, unsigned gpio_num)
{
	int idx;
	int tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	LOG_MESSAGE(KERN_INFO, "%s: %d\n", __func__, __LINE__);

	tmp = inb(priv->gpio_ioaddress + idx);

	return !!(tmp & (1 << gpio_num));
}

static int f81504_gpio_direction_in(struct gpio_chip *chip, unsigned gpio_num)
{
	int idx;
	u8 tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	mutex_lock(&priv->locker);

	/* set input mode */
	pci_read_config_byte(dev, 0xf8 + idx * 0x08, &tmp);
	pci_write_config_byte(dev, 0xf8 + idx * 0x08, tmp & ~(1 << gpio_num));

	mutex_unlock(&priv->locker);

	LOG_MESSAGE(KERN_INFO, "%s: gpio_num:%d, tmp:%d\n", __func__, gpio_num,
		    tmp);
	return 0;
}

static int f81504_gpio_direction_out(struct gpio_chip *chip, unsigned gpio_num,
				     int val)
{
	int idx;
	u8 tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	LOG_MESSAGE(KERN_INFO, "%s: %d\n", __func__, __LINE__);

	mutex_lock(&priv->locker);

	/* set output mode */
	pci_read_config_byte(dev, 0xf8 + idx * 0x08, &tmp);
	pci_write_config_byte(dev, 0xf8 + idx * 0x08, tmp | (1 << gpio_num));
#ifndef GPIO_MODE_OPENDRAIN
	/* set push pull mode */
	pci_read_config_byte(dev, 0xf8 + idx * 0x08 + 1, &tmp);
	pci_write_config_byte(dev, 0xf8 + idx * 0x08 + 1,
			      tmp | (1 << gpio_num));
#endif
	/* set output data */
	tmp = inb(priv->gpio_ioaddress + idx);

	if (val)
		outb(tmp | (1 << gpio_num), priv->gpio_ioaddress + idx);
	else
		outb(tmp & ~(1 << gpio_num), priv->gpio_ioaddress + idx);

	mutex_unlock(&priv->locker);

	return 0;
}

static void f81504_gpio_set(struct gpio_chip *chip, unsigned gpio_num, int val)
{
	LOG_MESSAGE(KERN_INFO, "%s: %d\n", __func__, __LINE__);
	f81504_gpio_direction_out(chip, gpio_num, val);
}

static int f81504_gpio_get_direction(struct gpio_chip *chip, unsigned offset)
{
	int idx;
	u8 tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	mutex_lock(&priv->locker);
	pci_read_config_byte(dev, 0xf8 + idx * 0x08, &tmp);
	mutex_unlock(&priv->locker);

	/* F81504 bit value 0: input, 1: output 
	 * but GPIOLIB getdirection return 0: output, 1: input
	 * so we need to do a revert !
	 */
	LOG_MESSAGE(KERN_INFO, "%s: idx:%d, tmp:%d, offset:%d\n", __func__,
		    idx, !!(tmp & (1 << offset)), offset);
	return !(tmp & (1 << offset));
}

static struct gpio_chip f81504_gpio_chip_templete = {
	.owner = THIS_MODULE,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	.get_direction = f81504_gpio_get_direction,
#endif
	.get = f81504_gpio_get,
	.direction_input = f81504_gpio_direction_in,
	.set = f81504_gpio_set,
	.direction_output = f81504_gpio_direction_out,
	.base = -1,
};
#endif

int f81534_check_baudrate(unsigned int baud, int *idx)
{
	int index;
	u32 quot[ARRAY_SIZE(baudrate_table)];
	u32 rem[ARRAY_SIZE(baudrate_table)];

	for (index = 0; index < ARRAY_SIZE(baudrate_table); ++index) {
		if (baud > baudrate_table[index])
			continue;

		quot[index] = DIV_ROUND_CLOSEST(baudrate_table[index], baud);
		rem[index] = baudrate_table[index] % baud;

		if (quot[index] && !rem[index]) {
			if (idx)
				*idx = index;
			return 0;
		}
	}

	return -1;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
void f81504_set_termios(struct uart_port *port, struct ktermios *termios,
			const struct ktermios *old)
#else
#if 1
void f81504_set_termios(struct uart_port *port, struct ktermios *termios,
			struct ktermios *old)
#else
// for centos stream kernel 5.14.0-390-el9
void f81504_set_termios(struct uart_port *port, struct ktermios *termios,
			const struct ktermios *old)
#endif
#endif
{
	struct pci_dev *dev = container_of(port->dev, struct pci_dev, dev);
	struct f81504_priv *priv = pci_get_drvdata(dev);
	unsigned int baud = tty_termios_baud_rate(termios);
	u8 tmp;
	int offset;
	int i;

	for (offset = 0; offset < MAX_FINTEK_SERIAL_PORT; ++offset)
		if ((priv->line_data[offset] != -1) &&
		    (priv->line_data[offset] == port->line))
			break;

	if (offset == MAX_FINTEK_SERIAL_PORT) {
		dev_err(port->dev, "%s failed!!\n", __func__);
		return;
	}

	dev_dbg(port->dev, "%s: offset: %d\n", __func__, offset);

	if (baud == 0)
		goto direct_term;

	dev_dbg(port->dev, "%s: 111 port->uartclk: %d, baud: %d\n", __func__,
		port->uartclk, baud);

	do {
		pci_read_config_byte(dev, 0x40 + offset * 0x08, &tmp);

		switch (baud) {
		case 134:
		case 110:
			port->uartclk = 115200 * 16;
			pci_write_config_byte(dev, 0x40 + offset * 0x08,
					      tmp & ~(0xc0));
			goto direct_term;
		}

		if (!f81534_check_baudrate(baud, &i)) {
			/* had optimize value */
			port->uartclk = baudrate_table[i] * 16;
			tmp = (tmp & ~(0xc0)) | clock_table[i];
			pci_write_config_byte(dev, 0x40 + offset * 0x08, tmp);
			dev_dbg(port->dev, "%s: optimized: %d\n", __func__, i);
			break;
		}

		if (old &&
		    !f81534_check_baudrate(tty_termios_baud_rate(old), NULL)) {
			baud = tty_termios_baud_rate(old);
			dev_dbg(port->dev, "%s: using old baud: %d\n",
				__func__, baud);
		} else {
			baud = 115200;
			dev_dbg(port->dev, "%s: force to baud: %d\n", __func__,
				baud);
		}

	} while (1);

	dev_dbg(port->dev, "%s: 222 port->uartclk: %d, baud: %d\n", __func__,
		port->uartclk, baud);

direct_term:
	serial8250_do_set_termios(port, termios, old);

#if 0
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)

	struct uart_8250_port {
		struct uart_port	port;
		struct timer_list	timer;		/* "no irq" timer */
		struct list_head	list;		/* ports on this IRQ */
		unsigned short		capabilities;	/* port capabilities */
		unsigned short		bugs;		/* port bugs */
		unsigned int		tx_loadsz;	/* transmit fifo load size */
		unsigned char		acr;
		unsigned char		ier;
		unsigned char		lcr;
		unsigned char		mcr;
		unsigned char		mcr_mask;	/* mask of user bits */
		unsigned char		mcr_force;	/* mask of forced bits */
		unsigned char		cur_iotype;	/* Running I/O type */
	
		/*
		 * Some bits in registers are cleared on a read, so they must
		 * be saved whenever the register is read but the bits will not
		 * be immediately processed.
		 */
#define LSR_SAVE_FLAGS UART_LSR_BRK_ERROR_BITS
		unsigned char		lsr_saved_flags;
#define MSR_SAVE_FLAGS UART_MSR_ANY_DELTA
		unsigned char		msr_saved_flags;
	
		/* 8250 specific callbacks */
		int			(*dl_read)(struct uart_8250_port *);
		void			(*dl_write)(struct uart_8250_port *, int);
	};
#endif
	struct uart_8250_port *up =
		container_of(port, struct uart_8250_port, port);

	up->mcr_force |= UART_MCR_LOOP;
	//printk(KERN_INFO "%s: up->mcr: %x\n", __func__, up->mcr);
	//printk(KERN_INFO "%s: up->mcr_mask: %x\n", __func__, up->mcr_mask);
	//serial8250_do_set_mctrl(port, TIOCM_LOOP);
#endif
}

#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)
static int f81504_rs485_config(struct uart_port *port,
			       struct ktermios *termios,
			       struct serial_rs485 *rs485)
#else
static int f81504_rs485_config(struct uart_port *port,
			       struct serial_rs485 *rs485)
#endif
{
	u8 setting = 0;
	int status;
	uintptr_t index = (uintptr_t)port->private_data;
	struct pci_dev *pci_dev = container_of(port->dev, struct pci_dev, dev);
	struct f81504_priv *priv = pci_get_drvdata(pci_dev);

	if (index < 0 || index >= MAX_FINTEK_SERIAL_PORT)
		return -EINVAL;

	status = pci_read_config_byte(pci_dev, 0x47 + 8 * index, &setting);
	if (status)
		return status;

	if (rs485->flags & SER_RS485_ENABLED) {
		setting |= BIT(4);

		if (rs485->flags & SER_RS485_RTS_ON_SEND)
			setting &= ~BIT(5);
		else
			setting |= BIT(5);
	} else {
		setting &= ~(BIT(4) | BIT(5));
	}

	status = pci_write_config_byte(pci_dev, 0x47 + 8 * index, setting);
	if (status)
		return status;

	port->rs485 = *rs485;
	priv->rs485[index] = *rs485;
	return 0;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)
static const struct serial_rs485 fintek_8250_rs485_supported = {
	.flags = SER_RS485_ENABLED | SER_RS485_RTS_ON_SEND |
		 SER_RS485_RTS_AFTER_SEND,
};
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
static int f81504_register_port(struct pci_dev *dev, int address,
				uintptr_t idx)
{
	struct uart_port port;
	memset(&port, 0, sizeof(port));
	port.iotype = UPIO_PORT;
	port.mapbase = 0;
	port.membase = NULL;
	port.regshift = 0;
	port.irq = dev->irq;
	port.flags |= UPF_SKIP_TEST | UPF_FIXED_TYPE | UPF_SHARE_IRQ;
	port.uartclk = 1843200;
	port.dev = &dev->dev;
	port.iobase = address;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	port.set_termios = f81504_set_termios;
#endif
	port.type = PORT_16550A;
	port.fifosize = 32;
	port.private_data = (void *)idx; /* save current idx */

	return serial8250_register_port(&port);
}
#else
static int f81504_register_port(struct pci_dev *dev, int address,
				uintptr_t idx)
{
	struct uart_8250_port port;

	memset(&port, 0, sizeof(port));

	port.port.iotype = UPIO_PORT;
	port.port.mapbase = 0;
	port.port.membase = NULL;
	port.port.regshift = 0;
	port.port.irq = dev->irq;
	port.port.flags |= UPF_SKIP_TEST | UPF_FIXED_TYPE | UPF_SHARE_IRQ;
	port.port.uartclk = 1843200;
	port.port.dev = &dev->dev;
	port.port.iobase = address;
	port.port.set_termios = f81504_set_termios;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	port.port.rs485_config = f81504_rs485_config;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)
	memcpy(&port.port.rs485_supported, &fintek_8250_rs485_supported,
	       sizeof(fintek_8250_rs485_supported));
#endif

	port.port.type = PORT_16550A;
	port.port.fifosize = 32;
	port.port.private_data = (void *)idx; /* save current idx */

	return serial8250_register_8250_port(&port);
}

#endif

static int f81504_fill_port_register(struct pci_dev *dev, int idx,
				     unsigned long *ioport)
{
	struct pci_dev *pdev = dev;
	u32 bar_data[3];
	u8 config_base, valid;
	u8 msr, lsr;
	int i;
	unsigned long iobase;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	u8 setting = 0;
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	struct serial_rs485 *rs485 = NULL;
#endif

	pci_read_config_dword(pdev, 0x24, &bar_data[0]);
	pci_read_config_dword(pdev, 0x20, &bar_data[1]);
	pci_read_config_dword(pdev, 0x1c, &bar_data[2]);

	if ((idx < 12) && (idx >= 0))
		config_base = 0x40 + idx * 0x08;
	else
		return -EINVAL;

	/* Calculate Real IO Port */
	iobase = (bar_data[idx / 4] & 0xffffffe0) + (idx % 4) * 8;

	/* Check is the port disabled by customer EEPROM */
	pci_read_config_byte(pdev, config_base + 0x00, &valid);
	//pr_info("idx: %d valid: %x\n", idx, valid);
	if ((valid & (BIT(2) | BIT(1))) == BIT(1)) {
		//pr_debug("%s: port idx: %d had disabled by EEPROM\n", __func__,
		//	 idx);
		return -ENODEV;
	}

	/* Enable UART I/O port */
	pci_write_config_byte(pdev, config_base + 0x00, 0x01);

	/* Select 128-byte FIFO and 8x FIFO threshold */
	pci_write_config_byte(pdev, config_base + 0x01, 0x33);

	/* LSB UART */
	pci_write_config_byte(pdev, config_base + 0x04, (u8)(iobase & 0xff));

	/* MSB UART */
	pci_write_config_byte(pdev, config_base + 0x05,
			      (u8)((iobase & 0xff00) >> 8));

	/* force RS232. 9Bits Disable */
	pci_write_config_byte(pdev, 0x47 + 8 * idx, 1);

	/* irq number, this usually fails, but the spec says to do it anyway. */
	pci_write_config_byte(pdev, config_base + 0x06, pdev->irq);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	rs485 = &priv->rs485[idx];

	if (rs485->flags & SER_RS485_ENABLED) {
		pci_read_config_byte(pdev, 0x40 + 8 * idx + 7, &setting);

		setting |= BIT(4);

		if (rs485->flags & SER_RS485_RTS_ON_SEND)
			setting &= ~BIT(5);
		else
			setting |= BIT(5);

		pci_write_config_byte(pdev, 0x40 + 8 * idx + 7, setting);
	}
#endif

	if (ioport != NULL)
		*ioport = iobase;

	/* Check Port exist */
	for (i = 0; i < 10; i++) {
		outb(0x10, iobase + UART_MCR);

		outb(UART_LCR_DLAB, iobase + UART_LCR);
		outb(0x01, iobase + UART_DLL);
		outb(0x00, iobase + UART_DLM);
		outb(0x00, iobase + UART_LCR);

		outb(0xc7, iobase + UART_FCR);
		lsr = inb(iobase + UART_LSR);
		msr = inb(iobase + UART_MSR);
		mdelay(1);

		outb(0x00, iobase + UART_MCR);
		mdelay(1); // delay

		lsr = inb(iobase + UART_LSR);
		msr = inb(iobase + UART_MSR);

		//pr_info("%s: 111port: %d lsr: %02x, msr: %02x\n", __func__,
		//	    idx, lsr, msr);

		/* Check H/W Port is exist */
		if ((lsr & UART_LSR_BI) && ((msr & 0xf0) == 0xf0))
			return -ENODEV;

		//pr_info("%s: 222port: %d lsr: %02x, msr: %02x\n", __func__,
		//	    idx, lsr, msr);
	}

	return 0;
}

static int f81504_is_gpio_set(int multi_function, int idx, int *set)
{
	int count;
	for (count = 0; count < MAX_FINTEK_SERIAL_GPIO; ++count) {
		if ((multi_function & (1 << count)) &&
		    (skip_port[count] == idx)) {
			if (set)
				*set = count;

			return 1;
		}
	}
	return 0;
}

static void f81504_init_gpio(struct pci_dev *dev)
{
	struct f81504_priv *priv = pci_get_drvdata(dev);
	u16 gpio_ioaddress;
	u8 tmp;

	pci_read_config_dword(dev, 0x18, &priv->gpio_ioaddress);
	priv->gpio_ioaddress &= 0xffffffe0;

	pci_read_config_word(dev, 0xf1, &gpio_ioaddress);

	if (gpio_ioaddress != priv->gpio_ioaddress) {
		pci_write_config_byte(dev, 0xf1,
				      (priv->gpio_ioaddress >> 0) & 0xff);
		pci_write_config_byte(dev, 0xf2,
				      (priv->gpio_ioaddress >> 8) & 0xff);
	}

	pci_read_config_byte(dev, 0xf0, &tmp);

	if (tmp != priv->multi_function) {
		/* Enable GPIO Port */
		pci_write_config_byte(dev, 0xf0, priv->multi_function);
	}
}

static int f81504_prepare_gpio(struct pci_dev *dev, int set)
{
#ifdef CONFIG_GPIOLIB
	struct f81504_priv *priv = pci_get_drvdata(dev);

	char *name = kmalloc(32, GFP_KERNEL);
	int rc;

	memcpy(&priv->f81504_gpio_chip[set], &f81504_gpio_chip_templete,
	       sizeof(f81504_gpio_chip_templete));

	sprintf(name, "%s%d", GPIO_DISPLAY_NAME, set);

	priv->f81504_gpio_chip[set].label = name;
	priv->f81504_gpio_chip[set].ngpio = 8;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	priv->f81504_gpio_chip[set].dev = &dev->dev;
#else
	priv->f81504_gpio_chip[set].parent = &dev->dev;
#endif
	rc = gpiochip_add(&priv->f81504_gpio_chip[set]);

	return rc;
#else
	return 0;
#endif
}

#if 0
static void time_start(struct timespec *start)
{
	getnstimeofday(start);
}

static unsigned long long time_delta(struct timespec *start)
{
#define SEC 1000000000LL
	long long secs, usecs;
	struct timespec stop;

	getnstimeofday(&stop);
	
	secs = stop.tv_sec - start->tv_sec;
	usecs = stop.tv_nsec - start->tv_nsec;
	if (usecs < 0) {
		secs--;
		usecs += SEC;
	}

	return secs * SEC + usecs;
}
#endif

#if EXTRA_GPIO_CHANGE_MODE
const static int m_accpet_mode_f81439[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };
const static int m_accpet_mode_f81435[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };
const static int m_accpet_mode_f81433[] = { 1, 2, 3, 4, 5, 7, -1 };

const static int *m_nAcceptArray[transceiver_type_invalid] = {
	[transceiver_type_f81433] = m_accpet_mode_f81433,
	[transceiver_type_f81435] = m_accpet_mode_f81435,
	[transceiver_type_f81439] = m_accpet_mode_f81439,
};

static int check_transceiver_mode(transceiver_type type, int mode)
{
	const int *tb;
	int i;

	if (type < transceiver_type_f81439 || type >= transceiver_type_invalid)
		return -1;

	tb = m_nAcceptArray[type];

	for (i = 0; tb[i] != -1; ++i) {
		if (mode == tb[i])
			return 0;
	}

	return -EINVAL;
}

static void writeb_delayed(u8 data, u8 __iomem *mem)
{
	writeb(data, mem);
	readb(mem);
}

static int __WriteI2CData_F81504_MEM__(struct pci_dev *pdev, unsigned int dev,
				       unsigned int addr, unsigned int data)
{
	struct f81504_priv *priv = pci_get_drvdata(pdev);
	int count;
	u8 tmp;

	/* Reset */
	writeb_delayed(0x80, priv->i2c_mem + 0x07);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x07], 0x80);

	count = 1000;
	while (--count) {
		tmp = readb(priv->i2c_mem + 0x07);
		if (!tmp)
			break;
	}

	if (!count) {
		//fprintf(stderr, "%s: reset I2C TIMEOUT\n", __func__);
		return -ETIMEDOUT;
	}

	/* device address */
	writeb_delayed(dev & 0xff, priv->i2c_mem + 0x02);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x02], dev & 0xff);

	/* enable WR_CMD_EN */
	writeb_delayed(0x03, priv->i2c_mem + 0x06);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x06], 0x03);

	/* clear busy flag */
	writeb_delayed(0x17, priv->i2c_mem + 0x01);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x01], 0x17);

	/* Set Reg addr */
	writeb_delayed(addr & 0xff, priv->i2c_mem + 0x03);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x03], addr & 0xff);

	/* Set Data */
	writeb_delayed(data & 0xff, priv->i2c_mem + 0x04);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x04], data & 0xff);

	/* Start Read, 0x02: EEPROM Write, 0x08: DBI Write?? */
	writeb_delayed(0x02, priv->i2c_mem + 0x00);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x00], 0x02);

	/* check not busy */
	count = 100;
	while (!(readb(priv->i2c_mem + 0x01) & 0x01) && count) {
		count -= 1;
		udelay(100);
	}

	/* check error */
	if (readb(priv->i2c_mem + 0x01) & (BIT(1) | BIT(2))) {
#if 0		
		if (readb(priv->i2c_mem + 0x01) & (BIT(1)))
			pr_info("%s: Write I2C NACK\n", __func__);

		if (readb(priv->i2c_mem + 0x01) & (BIT(2)))
			pr_info("%s: Write I2C TIMEOUT\n", __func__);
#endif
		return -ETIMEDOUT;
	}

	return 0;
}

static int __ReadI2CData_F81504_MEM__(struct pci_dev *pdev, unsigned int dev,
				      unsigned int addr, unsigned int *data)
{
	struct f81504_priv *priv = pci_get_drvdata(pdev);
	int count;
	u8 tmp;

	/* Reset */
	writeb_delayed(0x80, priv->i2c_mem + 0x07);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x07], 0x80);

	count = 1000;
	while (--count) {
		tmp = readb(priv->i2c_mem + 0x07);
		if (!tmp)
			break;
	}

	if (!count) {
		//fprintf(stderr, "%s: reset I2C TIMEOUT\n", __func__);
		return -ETIMEDOUT;
	}

	/* device address */
	writeb_delayed(dev & 0xff, priv->i2c_mem + 0x02);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x02], dev & 0xff);

	/* enable WR_CMD_EN */
	writeb_delayed(0x03, priv->i2c_mem + 0x06);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x06], 0x03);

	/* clear busy flag */
	writeb_delayed(0x17, priv->i2c_mem + 0x01);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x01], 0x17);

	/* Set Reg addr */
	writeb_delayed(addr & 0xff, priv->i2c_mem + 0x03);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x03], addr & 0xff);

	/* Start Read, 0x04: EEPROM Read, 0x10: DBI Read?? */
	writeb_delayed(0x04, priv->i2c_mem + 0x00);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x00], 0x04);

	/* check not busy */
	count = 100;
	while (!(readb(priv->i2c_mem + 0x01) & 0x01) && count) {
		count -= 1;
		udelay(100);
	}

	/* check error */
	if (readb(priv->i2c_mem + 0x01) & (BIT(1) | BIT(2))) {
#if 0		
		if (readb(priv->i2c_mem + 0x01) & (BIT(1)))
			pr_info("%s: Read I2C NACK\n", __func__);

		if (readb(priv->i2c_mem + 0x01) & (BIT(2)))
			pr_info("%s: Read I2C TIMEOUT\n", __func__);
#endif
		return -ETIMEDOUT;
	} else {
		/* Get Data */
		*data = readb(priv->i2c_mem + 0x04) & 0xff;
	}

	return 0;
}

static int __WriteI2CData_F81504_CONF__(struct pci_dev *pdev, unsigned int dev,
					unsigned int addr, unsigned int data)
{
	int delay = 300;
	int status = 0;
	int count;
	unsigned char tmp;

	/* Reset */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, 0x80);

	count = 1000;
	while (--count) {
		udelay(delay);
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, &tmp);
		if (!tmp)
			break;
	}

	if (!count) {
		//fprintf(stderr, "%s: reset I2C TIMEOUT\n", __func__);
		return -ETIMEDOUT;
	}

	/* device address */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x02 * 4, dev & 0xff);

	/* enable WR_CMD_EN */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x06 * 4, 0x03);

	/* clear busy flag */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, 0xff);

	/* Set Reg addr */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x03 * 4, addr & 0xff);

	/* Set data */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x04 * 4, data & 0xff);

	/* Start Read, 0x04: EEPROM Read, 0x10: DBI Read?? */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x00 * 4, 0x02);

	/* check not busy */
	//ptr[align_mem_offset + 0x01]

	//pci_read_config_byte(int addr,unsigned char * data)
	count = 1000;
	while (count) {
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, &tmp);
		if (tmp & 0x01)
			break;

		count -= 1;
	}

	/* check error */
	if (tmp & (BIT(1) | BIT(2))) {
#if 0		
		if (tmp & (BIT(1)))
			fprintf(stderr, "%s: Read I2C NACK\n", __func__);

		if (tmp & (BIT(2)))
			fprintf(stderr, "%s: Read I2C TIMEOUT\n", __func__);
#endif
		status = -ETIMEDOUT;
	}

	return status;
}

static int __ReadI2CData_F81504_CONF__(struct pci_dev *pdev, unsigned int dev,
				       unsigned int addr, unsigned int *data)
{
	int delay = 300;
	int status = 0;
	int count;
	unsigned char tmp;

	//pr_info("%s: dev: %x, addr: %x\n", __func__, dev, addr);

	/* Reset */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, 0x80);

	count = 1000;
	while (--count) {
		udelay(delay);
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, &tmp);
		if (!tmp)
			break;
	}

	if (!count) {
		//fprintf(stderr, "%s: reset I2C TIMEOUT\n", __func__);
		return -ETIMEDOUT;
	}

	/* device address */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x02 * 4, dev & 0xff);
	//msleep(delay);

	/* enable WR_CMD_EN */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x06 * 4, 0x03);
	//msleep(delay);

	/* clear busy flag */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, 0xff);
	//msleep(delay);

	/* Set Reg addr */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x03 * 4, addr & 0xff);
	//msleep(delay);

	/* Start Read, 0x04: EEPROM Read, 0x10: DBI Read?? */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x00 * 4, 0x04);
	//msleep(delay);

	/* check not busy */
	//ptr[align_mem_offset + 0x01]

	//pci_read_config_byte(int addr,unsigned char * data)
	count = 1000;
	while (count) {
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, &tmp);
		if (tmp & 0x01)
			break;

		count -= 1;
	}

	/* check error */
	if (tmp & (BIT(1) | BIT(2))) {
#if 0		
		if (tmp & (BIT(1)))
			fprintf(stderr, "%s: Read I2C NACK\n", __func__);

		if (tmp & (BIT(2)))
			fprintf(stderr, "%s: Read I2C TIMEOUT\n", __func__);
#endif
		status = -ETIMEDOUT;
	} else {
		/* Get Data */
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x04 * 4, &tmp);
		*data = tmp & 0xff;
	}

	//pr_info("%s: tmp: %x\n", __func__, tmp);

	return status;
}

static int __WriteI2CData_F81504__(struct pci_dev *pdev, unsigned int dev,
				   unsigned int addr, unsigned int data)
{
	struct f81504_priv *priv = pci_get_drvdata(pdev);

	if (priv->is_i2c_mem)
		return __WriteI2CData_F81504_MEM__(pdev, dev, addr, data);
	else
		return __WriteI2CData_F81504_CONF__(pdev, dev, addr, data);
}

static int __ReadI2CData_F81504__(struct pci_dev *pdev, unsigned int dev,
				  unsigned int addr, unsigned int *data)
{
	struct f81504_priv *priv = pci_get_drvdata(pdev);

	if (priv->is_i2c_mem)
		return __ReadI2CData_F81504_MEM__(pdev, dev, addr, data);
	else
		return __ReadI2CData_F81504_CONF__(pdev, dev, addr, data);
}

static int __WriteI2CData_Mask_F81504__(struct pci_dev *pdev, unsigned int dev,
					unsigned int addr, unsigned int mask,
					unsigned int data)
{
	unsigned int tmp;
	int status;

	status = __ReadI2CData_F81504__(pdev, dev, addr, &tmp);
	if (status)
		return status;

	tmp = (tmp & ~mask) | (data & mask);

	return __WriteI2CData_F81504__(pdev, dev, addr, tmp);
}

static void f81504_dump_f75111(struct pci_dev *dev)
{
	struct f81504_priv *priv = pci_get_drvdata(dev);
	u8 dump_tbl[] = { 0x10, 0x11, 0x12, 0x20, 0x21, 0x22 };
	int i, r, tmp;

	for (i = 0; i < ARRAY_SIZE(dump_tbl); ++i) {
		r = __ReadI2CData_F81504__(dev, priv->gpio_dev_id, dump_tbl[i],
					   &tmp);
		pr_info("%s: reg: %02x, data: %02x\n", __func__, dump_tbl[i],
			tmp);
	}
}

static int get_gpio_switch_devid(struct pci_dev *pdev, int *devid)
{
	unsigned int tmp;
	int status;

	*devid = 0x9c;
	status = __ReadI2CData_F81504__(pdev, *devid, 0x5d, &tmp);
	if (status) {
		*devid = 0x6e;

		status = __ReadI2CData_F81504__(pdev, *devid, 0x5d, &tmp);
		if (status) {
			dev_info(
				&pdev->dev,
				"Cant found any GPIO IC, no set_gpio support (1)\n");
			return -ENODEV;
		}
	}

	//pr_info("%s: tmp: %x\n", __func__, tmp);

	if (tmp != 0x19) {
		dev_info(&pdev->dev,
			 "Cant found any GPIO IC, no set_gpio support (2)\n");
		return -ENODEV;
	}

	dev_info(&pdev->dev, "found GPIO IC(%02x), set_gpio supported\n",
		 *devid);

	return 0;
}

static int get_transceiver_type(struct pci_dev *pdev, int dev_id, int idx,
				transceiver_type *type)
{
	unsigned int tmp;
	int status;

	/* read F75111 GPIO 17 for all option: 433/439 or 435 */
	status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10, BIT(7), 0);
	status = __ReadI2CData_F81504__(pdev, dev_id, 0x12, &tmp);
	if (status) {
		pr_info("%s: Read all transceiver type failed\n", __func__);
		return status;
	}

	if (tmp & BIT(7)) {
		status = __ReadI2CData_F81504__(pdev, dev_id, 0x42, &tmp);
		if (status) {
			pr_info("%s: Read one transceiver type failed\n",
				__func__);
			return status;
		}

		if (tmp & BIT(idx))
			*type = transceiver_type_f81439;
		else
			*type = transceiver_type_f81433;

	} else {
		*type = transceiver_type_f81435;
	}

	return 0;
}

static int is_hw_fix_mode(struct pci_dev *pdev, int dev_id, int idx,
			  int *is_fix)
{
	int status, tmp, mode;

	switch (idx) {
	case 0:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 0, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x12, &tmp);
		mode = (tmp >> 0) & 0x07;
		break;
	case 1:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 4, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x12, &tmp);
		mode = (tmp >> 4) & 0x07;
		break;
	case 2:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 0, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x22, &tmp);
		mode = (tmp >> 0) & 0x07;
		break;
	case 3:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 4, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x22, &tmp);
		mode = (tmp >> 4) & 0x07;
		break;
	default:
		return -EINVAL;
	}

	if (mode != 0x07)
		*is_fix = 1;
	else
		*is_fix = 0;

	return status;
}

static int change_gpio(struct pci_dev *pdev, int dev_id, int idx, int mode)
{
	unsigned int tmp = 0;
	int status;

	if (mode & BIT(0))
		tmp |= BIT(2);
	if (mode & BIT(1))
		tmp |= BIT(1);
	if (mode & BIT(2))
		tmp |= BIT(0);

	switch (idx) {
	case 0:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 0, 0x07 << 0);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x11,
						       0x07 << 0, tmp << 0);
		break;
	case 1:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 4, 0x07 << 4);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x11,
						       0x07 << 4, tmp << 4);
		break;
	case 2:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 0, 0x07 << 0);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x21,
						       0x07 << 0, tmp << 0);
		break;
	case 3:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 4, 0x07 << 4);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x21,
						       0x07 << 4, tmp << 4);
		break;
	default:
		status = -1;
		break;
	}

	return status;
}

static ssize_t gpio_ctrl_store(struct device *dev,
			       struct device_attribute *attr, const char *buf,
			       size_t count)
{
	struct f81504_priv *priv;
	struct pci_dev *pdev;
	int i, r;
	unsigned int data[2];
	u8 *p;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	for (i = 0; i < 2; ++i) {
		p = strsep((char **)&buf, " ");
		if (!p) {
			return -EINVAL;
		}

		r = kstrtoint(p, 0, &data[i]);
		if (r) {
			return r;
		}

		if (data[i] >= 256 || data[i] < 0) {
			return -EINVAL;
		}
	}

	if (data[0] >= 4)
		return -EINVAL;

	if (data[1] >= 8)
		return -EINVAL;

	mutex_lock(&priv->locker);

	if (priv->gpio_mode[data[0]] == 0xff) {
		pr_info("%s: cant change hw_fix port: %d\n", __func__,
			data[0]);
		r = -EPERM;
		goto end;
	}

	r = check_transceiver_mode(priv->transceiver_type[data[0]], data[1]);
	if (r)
		goto end;

	priv->gpio_mode[data[0]] = data[1];

	//pr_info("%s: idx: %d, data: %d\n", __func__, data[0], data[1]);
	r = change_gpio(pdev, priv->gpio_dev_id, data[0], data[1]);
end:
	mutex_unlock(&priv->locker);

	//f81504_dump_f75111(pdev);

	if (!r)
		return count;

	pr_info("%s: r: %d\n", __func__, r);
	return r;
}

static ssize_t gpio_ctrl_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct f81504_priv *priv;
	struct pci_dev *pdev;
	int i, len = 0;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	mutex_lock(&priv->locker);

	for (i = 0; i < 4; ++i)
		len += sprintf(&buf[len], "%d \n", priv->gpio_mode[i]);

	mutex_unlock(&priv->locker);

	return len;
}

static DEVICE_ATTR_RW(gpio_ctrl);

static ssize_t transceiver_show(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct f81504_priv *priv;
	struct pci_dev *pdev;
	int i, len = 0;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	for (i = 0; i < 4; ++i)
		len += sprintf(&buf[len], "%d\n", priv->transceiver_type[i]);

	return len;
}

static DEVICE_ATTR_RO(transceiver);

static ssize_t f75111_show(struct device *dev, struct device_attribute *attr,
			   char *buf)
{
	struct f81504_priv *priv;
	struct pci_dev *pdev;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	f81504_dump_f75111(pdev);

	return 0;
}

static DEVICE_ATTR_RO(f75111);

static int f81504_init_f75111_info(struct pci_dev *dev, bool is_first)
{
	struct f81504_priv *priv = pci_get_drvdata(dev);
	uint32_t i2c_mem, i;
	int r, is_fix;
	u8 tmp;

	mutex_lock(&priv->locker);

	pci_read_config_byte(dev, 0xb9, &tmp);
	tmp |= 0xe0;
	pci_write_config_byte(dev, 0xb9, tmp);

	tmp = 0xa0;
	pci_write_config_byte(dev, F81504_I2C_CONF + 0x02 * 4, tmp);

	priv->is_i2c_mem = false;
	for (i = 0; i < 4; ++i) {
		pci_read_config_byte(dev, F81504_I2C_CONF + i + 0x02 * 4,
				     &tmp);
		if (tmp != 0xa0) {
			priv->is_i2c_mem = true;
			break;
		}
	}

	//pr_info("%s: priv->is_i2c_mem: %d\n", __func__, priv->is_i2c_mem);
	//pr_info("%s: tmp: %x\n", __func__, tmp);

	if (priv->is_i2c_mem) {
		pci_read_config_dword(dev, 0x14, &i2c_mem);
		pci_write_config_dword(dev, 0xb4, i2c_mem);

		priv->i2c_mem = pci_ioremap_bar(dev, 1);
	}

	r = get_gpio_switch_devid(dev, &priv->gpio_dev_id);
	if (r) {
		priv->gpio_dev_id = NO_GPIO_DEV;
		//pr_err("%s: get_gpio_switch_devid err: %d\n", __func__, r);
		goto end;
	}

#if 0
	unsigned int data[2];

	//test
	pr_info("%s: dev_id: %x\n", __func__, priv->gpio_dev_id);

	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x15, 0xff);
	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x25, 0xff);
	
	//f81504_dump_f75111(dev);
	udelay(100);

	__ReadI2CData_F81504__(dev, priv->gpio_dev_id, 0x12, &data[0]);
	__ReadI2CData_F81504__(dev, priv->gpio_dev_id, 0x22, &data[1]);

	pr_info("%s: %x %x\n", __func__, data[0], data[1]);

	pr_info("%s: M0_1: %d\n", __func__, !!(data[0] & BIT(0)));
	pr_info("%s: M1_1: %d\n", __func__, !!(data[0] & BIT(1)));
	pr_info("%s: M2_1: %d\n", __func__, !!(data[0] & BIT(2)));
	
	pr_info("%s: M0_2: %d\n", __func__, !!(data[0] & BIT(4)));
	pr_info("%s: M1_2: %d\n", __func__, !!(data[0] & BIT(5)));
	pr_info("%s: M2_2: %d\n", __func__, !!(data[0] & BIT(6)));

	pr_info("%s: M0_3: %d\n", __func__, !!(data[1] & BIT(0)));
	pr_info("%s: M1_3: %d\n", __func__, !!(data[1] & BIT(1)));
	pr_info("%s: M2_3: %d\n", __func__, !!(data[1] & BIT(2)));

	pr_info("%s: M0_4: %d\n", __func__, !!(data[1] & BIT(4)));
	pr_info("%s: M1_4: %d\n", __func__, !!(data[1] & BIT(5)));
	pr_info("%s: M2_4: %d\n", __func__, !!(data[1] & BIT(6)));
	
	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x15, 0x00);
	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x25, 0x00);

	goto end;
#endif

	for (i = 0; i < 4; ++i) {
		r = get_transceiver_type(dev, priv->gpio_dev_id, i,
					 &priv->transceiver_type[i]);
		if (r) {
			pr_err("%s: get_transceiver_type err: %d\n", __func__,
			       r);
			goto end;
		}

		r = is_hw_fix_mode(dev, priv->gpio_dev_id, i, &is_fix);
		if (r) {
			pr_err("%s: is_hw_fix_mode err: %d\n", __func__, r);
			goto end;
		}

		if (is_first)
			priv->gpio_mode[i] = 1;

		if (!is_fix) {
			r = change_gpio(dev, priv->gpio_dev_id, i,
					priv->gpio_mode[i]);
			if (r) {
				pr_err("%s: change_gpio err: %d\n", __func__,
				       r);
				goto end;
			}
		} else {
			priv->gpio_mode[i] = 0xff;
		}
	}

end:
	mutex_unlock(&priv->locker);
	return r;
}

#if 0
static int f81504_reapply_f75111_info(struct pci_dev *dev)
{
	struct f81504_priv *priv = pci_get_drvdata(dev);
	int i, r, is_fix;

	mutex_lock(&priv->locker);

	for (i = 0; i < 4; ++i) {
		r = is_hw_fix_mode(dev, priv->gpio_dev_id, i, &is_fix);
		if (r)
			goto end;

		r = change_gpio(dev, priv->gpio_dev_id, i, priv->gpio_mode[i]);
		if (r)
			goto end;

		//pr_info("%s: change idx: %d, mode: %d\n", __func__, i, priv->gpio_mode[i]);
	}

end:
	mutex_unlock(&priv->locker);
	return 0;
}
#endif

static int f81504_prepare_mode_pin(struct pci_dev *dev)
{
	int r;

	r = f81504_init_f75111_info(dev, true);
	if (r)
		return r;

	r = device_create_file(&dev->dev, &dev_attr_f75111);
	if (r)
		goto attr_f75111_err;

	r = device_create_file(&dev->dev, &dev_attr_gpio_ctrl);
	if (r)
		goto attr_gpio_ctrl_err;

	r = device_create_file(&dev->dev, &dev_attr_transceiver);
	if (r)
		goto attr_transceiver_err;

	return 0;

attr_transceiver_err:
	device_remove_file(&dev->dev, &dev_attr_gpio_ctrl);
attr_gpio_ctrl_err:
	device_remove_file(&dev->dev, &dev_attr_f75111);
attr_f75111_err:

	return r;
}

static void f81504_remove_mode_pin(struct pci_dev *dev)
{
	struct f81504_priv *priv = pci_get_drvdata(dev);

	if (priv->gpio_dev_id != NO_GPIO_DEV) {
		device_remove_file(&dev->dev, &dev_attr_f75111);
		device_remove_file(&dev->dev, &dev_attr_transceiver);
		device_remove_file(&dev->dev, &dev_attr_gpio_ctrl);
	}

	iounmap(priv->i2c_mem);
}
#else
static int f81504_init_f75111_info(struct pci_dev *dev, , bool is_first)
{
	return 0;
}

#if 0
static int f81504_reapply_f75111_info(struct pci_dev *dev)
{
	return 0;
}
#endif

static int f81504_prepare_mode_pin(struct pci_dev *dev)
{
	return 0;
}

static void f81504_remove_mode_pin(struct pci_dev *dev)
{
}
#endif

static int f81504_probe(struct pci_dev *dev,
			const struct pci_device_id *dev_id)
{
	int idx;
	int rc;
	int set;
	u8 sirq;
	struct f81504_priv *priv = NULL;
	unsigned long iobase = 0;

	rc = pci_enable_device(dev);
	if (rc)
		return rc;

	priv = kzalloc(sizeof(struct f81504_priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	pci_set_drvdata(dev, priv);
	mutex_init(&priv->locker);

#if 1
	/* check SERIRQ trap to GPIO */
	//pci_read_config_byte(dev, 0xb8, &old_sirq);
	pci_write_config_byte(dev, 0xb8, 0x01);

	udelay(100);

	pci_read_config_byte(dev, 0xb8, &sirq);
	if ((sirq & BIT(2)) == 0) {
		/* force into gpio mode */
		pci_write_config_byte(dev, 0xf3, 0x00);

		dev_info(&dev->dev, "SIRQ force gpio mode\n");
	}

	dev_info(&dev->dev, "%s: sirq: %x\n", __func__, sirq);
	pci_write_config_byte(dev, 0xb8, 0x06);
#endif

#ifdef FORCE_INTO_GPIO_MODE
	pci_write_config_byte(dev, 0xf3, 0x00);
#endif

	pci_read_config_byte(dev, 0xf3, &priv->multi_function);

	/* max multi function with 0x3f - 6devices */
	priv->multi_function = (~priv->multi_function) & 0x3f;

	dev_info(&dev->dev, "%s: Multifunction inv-Flag:%d\n", __func__,
		 (int)priv->multi_function);

	for (idx = 0; idx < MAX_FINTEK_SERIAL_PORT; ++idx) {
		dev_dbg(&dev->dev, "port %d change to GPIO\n", idx);
		priv->line_data[idx] = MULTI_FUNCTION_PORT_FLAG;
	}

	f81504_init_gpio(dev);

	rc = f81504_prepare_mode_pin(dev);
	if (rc && rc != -ENODEV) {
		kfree(priv);
		return rc;
	}

	for (idx = 0; idx < dev_id->driver_data; ++idx) {
		if (disable_port & BIT(idx)) {
			pr_debug("%s: disable: %d\n", __func__, idx);
			continue;
		}

		if (f81504_is_gpio_set(priv->multi_function, idx, &set)) {
			rc = f81504_prepare_gpio(dev, set);
			if (rc) {
				LOG_MESSAGE(KERN_INFO,
					    "%s: gpiochip_add failed:%d\n",
					    __func__, rc);
			}
		} else {
			rc = f81504_fill_port_register(dev, idx, &iobase);
			if (rc < 0) {
				if (rc == -ENODEV) /* port disbled */
					continue;

				printk(KERN_WARNING "port idx: %d,failed!!\n",
				       idx);
				continue;
			}

			priv->line_data[idx] =
				f81504_register_port(dev, iobase, idx);

			dev_dbg(&dev->dev, "idx: %d\n", idx);
			dev_dbg(&dev->dev, "register with: %d\n",
				priv->line_data[idx]);
		}
	}

	if (dev_id->driver_data != 8)
		return 0;

	/* if F81508, 2multi + 4gpio sets need check */
	for (idx = dev_id->driver_data; idx < MAX_FINTEK_SERIAL_PORT; ++idx) {
		if (!f81504_is_gpio_set(priv->multi_function, idx, &set))
			continue;

		rc = f81504_prepare_gpio(dev, set);
		if (rc) {
			LOG_MESSAGE(KERN_INFO, "%s: gpiochip_add failed:%d\n",
				    __func__, rc);
		}
	}

	dev->dev_flags |= PCI_DEV_FLAGS_NO_D3;
	return 0;
}

static void f81504_remove(struct pci_dev *dev)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 18, 0)
	int rc;
#endif
	int i;

	LOG_MESSAGE(KERN_INFO, "f81504_remove\n");

	if (priv) {
		for (i = 0; i < MAX_FINTEK_SERIAL_PORT; ++i)
			if (priv->line_data[i] > 0) {
				serial8250_unregister_port(priv->line_data[i]);
			}

#ifdef CONFIG_GPIOLIB
		for (i = 0; i < MAX_FINTEK_SERIAL_GPIO; ++i) {
			if (priv->f81504_gpio_chip[i].label) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)
				gpiochip_remove(&priv->f81504_gpio_chip[i]);
#else
				rc = gpiochip_remove(
					&priv->f81504_gpio_chip[i]);
				if (rc)
					LOG_MESSAGE(KERN_INFO,
						    "%s: gpiochip_remove:%d\n",
						    __func__, rc);
#endif
				kfree(priv->f81504_gpio_chip[i].label);
			}
		}
#endif

		kfree(priv);
	}

	f81504_remove_mode_pin(dev);

	pci_set_drvdata(dev, NULL);
	pci_disable_device(dev);
}

static void f81504_suspend_save_data(struct pci_dev *dev)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	int i;

	for (i = 0; i < MAX_FINTEK_SERIAL_PORT; ++i)
		pci_read_config_byte(dev, 0x40 + 8 * i + 7,
				     &priv->rs485_reg[i]);
}

static void f81504_resume_restore_data(struct pci_dev *dev)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	int i;

	for (i = 0; i < MAX_FINTEK_SERIAL_PORT; ++i)
		pci_write_config_byte(dev, 0x40 + 8 * i + 7,
				      priv->rs485_reg[i]);
}

static int f81504_suspend(struct pci_dev *dev, pm_message_t state)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	int i;

	pci_save_state(dev);
	f81504_suspend_save_data(dev);
	pci_set_power_state(dev, pci_choose_state(dev, state));

	if (!priv)
		return -ENODEV;

	for (i = 0; i < MAX_FINTEK_SERIAL_PORT; ++i)
		if (priv->line_data[i] > 0)
			serial8250_suspend_port(priv->line_data[i]);

	return 0;
}

static int f81504_resume(struct pci_dev *dev)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	int i, devices;
	int rc;

	pci_set_power_state(dev, PCI_D0);
	pci_restore_state(dev);

	if (!priv)
		return -ENODEV;

	switch (dev->device) {
	case 0x1104: /* 4 ports */
	case 0x1108: /* 8 ports */
		devices = dev->device & 0xff;
		break;
	case 0x1112: /* 12 ports */
		devices = 12;
		break;
	default:
		return -EINVAL;
	}

	f81504_init_gpio(dev);

	for (i = 0; i < devices; ++i) {
		if (priv->line_data[i] > 0) {
			rc = f81504_fill_port_register(dev, i, NULL);
			if (rc < 0) {
				printk(KERN_INFO "%s: idx: %d failed\n",
				       __func__, i);
				//continue;
			}

			serial8250_resume_port(priv->line_data[i]);
		}
	}

	f81504_resume_restore_data(dev);

#if 0
	rc = f81504_init_f75111_info(dev);
	if (rc)
		return rc;

	return f81504_reapply_f75111_info(dev);
#else
	return f81504_init_f75111_info(dev, false);
#endif
}

static const struct pci_device_id f81504_dev_table[] = {
	/* Fintek PCI serial cards */
	{ PCI_DEVICE(0x1c29, 0x1104), .driver_data = 4 },
	{ PCI_DEVICE(0x1c29, 0x1108), .driver_data = 8 },
	{ PCI_DEVICE(0x1c29, 0x1112), .driver_data = 12 },
	{}
};

MODULE_DEVICE_TABLE(pci, f81504_dev_table);

static struct pci_driver f81504_driver = {
	.name = "f81504_pci2uart-4-8-12",
	.probe = f81504_probe,
	.remove = f81504_remove,
#ifdef CONFIG_PM
	.suspend = f81504_suspend,
	.resume = f81504_resume,
#endif
	.id_table = f81504_dev_table,
};

static __init int f81504_init(void)
{
	pr_info("F81504 series Driver version: %s\n", F81504_DRIVER_VER);

	return pci_register_driver(&f81504_driver);
}

static __exit void f81504_exit(void)
{
	pci_unregister_driver(&f81504_driver);
}

module_init(f81504_init);
module_exit(f81504_exit);

MODULE_DESCRIPTION(DRIVER_NAME);
MODULE_AUTHOR("Peter Hong <Peter_Hong@fintek.com.tw>");
MODULE_LICENSE("GPL");
